from dknovautils import commons, iprint_debug, iprint_warn, AT  # type:ignore
from dknovautils.commons import *

import urllib3

http = urllib3.PoolManager(timeout=5.0)


class DkNetwork:

    @staticmethod
    def get_url_simple_str(url: str) -> str | None:
        try:
            r = http.request("GET", url)
            # print(json.loads(r.data.decode('utf-8')))
            s = r.data.decode("utf-8")
            assert isinstance(s, str)
            return s
        except Exception as e:
            iprint_debug(e)
            return None



from __future__ import annotations

from dknovautils.commons import *

import pathlib


_debug = False


class DkFile(object):
    def __init__(self, pathstr: str) -> None:
        self.pathstr = str(pathstr)

        self.path = Path(pathstr)

    def __key(self) -> Any:
        return self.path

    def __hash__(self) -> int:
        return hash(self.__key())

    def __eq__(self, other: Any) -> Any:
        if isinstance(other, DkFile):
            return self.__key() == other.__key()
        return NotImplemented

    @property
    def basename(self) -> str:
        return os.path.basename(self.path)

    @property
    def filesize(self) -> int:
        return getsize(self.path)

    @property
    def dirname(self) -> str:
        return os.path.dirname(self.path)

    @property
    def extension(self) -> str:
        """包括点号 比如 .txt 如果没有扩展名 则返回空串"""
        # function to return the file extension
        file_extension = self.path.suffix
        return file_extension

    def exists(self) -> bool:
        return self.path.exists()

    def is_file(self) -> bool:
        return self.path.is_file()

    def is_dir(self) -> bool:
        return self.path.is_dir()

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return str(self.path)

    @staticmethod
    def clear_dir(folder: str) -> None:
        """

        https://stackoverflow.com/questions/185936/how-to-delete-the-contents-of-a-folder

        """
        import os, shutil

        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                iprint("Failed to delete %s. Reason: %s" % (file_path, e))

    @staticmethod
    def file_md5(f: str, md5Cache: Dict[str, str] | None = None) -> str:
        # todo 更换为合理的计算方式

        if md5Cache is None:
            md5Cache = {}

        if f in md5Cache:
            r = md5Cache[f]
            assert len(r) == 32
            return r
        else:
            iprint_debug(f"gen md5 {f}")
            bs = Path(f).read_bytes()
            md5 = hashlib.md5()
            # md5.update('how to use md5 in python hashlib?'.encode('utf-8'))
            md5.update(bs)
            r = md5.hexdigest().lower()
            assert len(r) == 32

            md5Cache[f] = r
            return r

    @staticmethod
    def listdir(d: str) -> List[DkFile]:
        fs = [DkFile(join(d, f)) for f in os.listdir(d)]
        # fs=[DkFile(f) for f in fs]
        return fs

    @staticmethod
    def file_sha1(f: str) -> str:
        import hashlib

        bs = Path(f).read_bytes()
        md5 = hashlib.sha1()
        # md5.update('how to use md5 in python hashlib?'.encode('utf-8'))
        md5.update(bs)
        r = md5.hexdigest().lower()
        assert len(r) == 32
        return r

    @staticmethod
    def fill_disk_simple(
        dir: str = ".",
        *,
        size: int = 1024 * 1024 * 100,
        pre: str = "tmp",
        verbose: bool = False,
    ) -> None:
        import numpy as np

        """
        在目录下创建文件. 文件大小约为10M. 创建之后马上读出检查内容是否正确. 用来对硬盘进行简单的检查.
        直到写满磁盘为止, [Errno 28] No space left on device 最后退出.
        
        mkdir tmp
        python3.10 -c "from dknovautils import *; DkFile.fill_disk_simple(dir='./tmp')"

        """
        if isinstance(size, int):
            assert size >= 1024 * 1024 * 10
            N = size // len(bs)
        else:
            raise Exception("unimplemented")

        pre = "tmp"
        bse = b"a"
        bs = np.frombuffer(bse * (1024 * 1024), dtype=np.uint8)

        def fpath(fid: int) -> Path:
            return Path(f"{dir}/{pre}{fid:08d}")

        print("begin write file")
        fid = 1
        while True:
            file = fpath(fid)
            if verbose:
                print(f"write file {file}")
            try:
                with file.open(mode="wb") as f:
                    for i in range(N):
                        f.write(bs)
                    f.flush()
                fid += 1
            except Exception as e:
                print(e)
                break

        print("begin verify")
        fid, maxfid = 1, fid
        while True:
            if fid >= maxfid:
                break
            file = fpath(fid)
            if verbose:
                print(f"verify file {file}")
            try:
                with file.open(mode="rb") as f:
                    r = f.read(len(bs))
                    bs2 = np.frombuffer(r, dtype=np.uint8)
                    if len(bs2) == len(bs):
                        assert np.all(bs2 == bs), "err56373 verify error"
                    else:
                        condition = bs2 == bs[0]
                        assert np.all(condition), "err56374 verify error"
                fid += 1
            except Exception as e:
                print(e)
                break
        print("all end")


class DkPyFiles(object):
    pass


if __name__ == "__main__":
    print("OK")



from __future__ import annotations

# from typing import Callable, Iterator, Set
# from pathlib import Path
# import re

# from typing import List, Dict, Tuple, Union, Any


# from dknovautils.dk_imports import *

# 导入module避免某些循环import的问题
from dknovautils import commons, iprint_debug, iprint_warn, iprint, AT
from dknovautils.commons import *


class DkLangEtcUtils(object):
    _debug = True

    @classmethod
    def f_remove_comments_v1(
        cls,
        py: str,
        *,
        _removeAstSentence: bool = False,
        _removeATVast: bool = False,
        _removeSingleQuoteMuilti: bool = False,
    ) -> str:
        """
        行尾的注释需要保留

        缺省情况下 单引号多行文本会被删除

        _removeSingleQuoteMuilti 参数 可以切换是单引号还是双引号


        """

        # 可以在py文件中添加如下字符串阻止移除多行注释 分开写是让此文件本身不会被错误识别
        KEEP = "__keep_" + "quotes__"

        assert py is not None, "err5242"

        # 注意ln已经有末尾的回车符 最好替换成\n方便后续处理 rstrip 将去掉末尾的回车符

        lns = "\n".join(ln.rstrip() for ln in py.splitlines())

        keep_quotes = KEEP in lns

        if cls._debug:
            print(lns)
            print("=====")

        # 去掉单行注释
        re_single_a = r"\n\s*#.*"
        pattern = re.compile(re_single_a)
        lns = re.sub(pattern, "", lns)  # 查找匹配 然后替换

        """
        是匹配单引号还是匹配双引号？
        
        写注释习惯用得最多的是双引号。所以还是缺省删除双引号。


        """

        # 首行开始的文本不会被替换掉，因为前方没有回车符
        if _removeSingleQuoteMuilti:
            fr_multi = r"(?s)'''.*?'''"
        else:
            fr_multi = r'(?s)""".*?"""'
        pattern = re.compile(fr_multi)
        if not keep_quotes:
            lns = re.sub(pattern, "pass", lns)  # 查找匹配 然后替换

        # 暂时关闭
        # _removeAstSentence = False
        if _removeAstSentence:
            re_single_a = r"\n\s*assert\s.+"
            pattern = re.compile(re_single_a)
            lns = re.sub(pattern, "", lns)

        # 暂时关闭的

        if _removeATVast:
            re_single_a = r"\n\s*AT.vassert_\(.+"
            pattern = re.compile(re_single_a)
            lns = re.sub(pattern, "", lns)

        if cls._debug:
            print(lns)

        return lns

    langWordSet = r"[a-zA-Z0-9\_\u4e00-\u9fa5]"

    # langNotWordSet = r'[^a-zA-Z0-9_\u4e00-\u9fa5]'

    @classmethod
    def find_words(
        cls, texts: List[str], keyFun: Callable[[str], Any], *, minLen: int = 6
    ) -> Set[str]:
        assert keyFun is not None
        rs = set()
        langWord = f"{cls.langWordSet}+"
        pWord = re.compile(langWord)

        for text in texts:
            assert len(text) >= 0  # and len(keyStr) >= 1
            ws = pWord.findall(text)
            ws = [x for x in ws if len(x) >= minLen and keyFun(x)]
            rs.update(ws)

        iprint(f"rs:{rs}")

        return rs

    @classmethod
    def sub_word(cls, texts: List[str], *, fromWord: str, toWord: str) -> List[str]:
        langWord = f"{cls.langWordSet}+"
        pWord = re.compile(langWord)
        AT.assert_(re.fullmatch(pWord, fromWord) is not None)
        AT.assert_(re.fullmatch(pWord, toWord) is not None)

        iprint(f"fromw:{fromWord} tow:{toWord}")

        r33str = f"""
        (?<=[^a-zA-Z0-9\_\u4e00-\u9fa5])
        {fromWord}
        (?=[^a-zA-Z0-9\_\u4e00-\u9fa5])    
        
        """
        r33 = re.compile(r33str, re.X)

        def fa() -> Iterator[str]:
            for text in texts:
                assert isinstance(text, str)
                yield re.sub(r33, toWord, text)

        rs = list(fa())

        return rs

    @classmethod
    def f_replace_text_simplea(cls, filepath: str, fcb: Callable[[str], str]) -> None:
        # 最简单的方式 直接将文件内容读出文本 utf8编码的文本
        # assert filepath and fcb

        # 如果不设置编码 缺省为系统缺省编码gbk
        text = Path(filepath).read_text(encoding="utf8")
        # print(text)
        text = fcb(text)
        AT.assert_(text is not None)
        Path(filepath).write_text(text, encoding="utf8")

    @classmethod
    def handle_files_100(cls, fs: List[Path], keyFun: Callable[[str], Any]) -> None:
        """
        即使在字符串常量中 也会替换的
        """
        texts = [fp.read_text() for fp in fs]
        iprint(f"files: {len(fs)}")

        assert sorted(["ab", "ba", "caa"], key=lambda x: x[::-1]) == ["caa", "ba", "ab"]

        ws = list(DkLangEtcUtils.find_words(texts, keyFun))

        if len(ws) <= 1:
            iprint_warn(f"err9667")
            return

        ws = sorted(ws, key=lambda x: x[::-1])

        tmpf = f"K_STS_K_{AT.fepochMillis()}"

        texts = DkLangEtcUtils.sub_word(texts, fromWord=ws[0], toWord=tmpf)
        subs = list(zip(ws, [*ws[1:], ws[0]]))
        for i, w in enumerate(subs[:-1]):
            tow, fromw = w

            texts = DkLangEtcUtils.sub_word(texts, fromWord=fromw, toWord=tow)

        texts = DkLangEtcUtils.sub_word(texts, fromWord=tmpf, toWord=ws[-1])

        for i, file in enumerate(fs):
            file.write_text(texts[i])


def _f_main() -> None:
    rPyId = r"([a-zA-Z0-9_\u4e00-\u9fa5]+)"

    pattern = re.compile(rPyId)  # 查找数字
    ta = (
        "runoob 123 google 456  \n 大家比里好 和*比/来.(这+里)里}里{"
        """
    xyz
    def 里()
        里好=abc
        里2=ttt
        x.里=xyz

    
    """
    )
    print(ta.splitlines())
    result1 = pattern.findall(ta)
    result2 = pattern.findall("run88oob123google456")

    rs = DkLangEtcUtils.find_words([ta, ta], lambda x: 2, minLen=2)
    iprint(rs)

    print(result1)
    print(result2)

    r22 = r"[^a-zA-Z0-9_\u4e00-\u9fa5]+"

    r33str = """
    (?<=[^a-zA-Z0-9_\u4e00-\u9fa5])
    #(?:[^a-zA-Z0-9_\u4e00-\u9fa5]+)
    里
    (?=[^a-zA-Z0-9_\u4e00-\u9fa5])    
    #(?:[^a-zA-Z0-9_\u4e00-\u9fa5]+)
    
    """

    ra = f"{r22}abc{r22}"
    r33 = re.compile(r33str, re.X)
    ta2 = re.sub(r33, "东西", ta)
    # print(ta2)

    ta3 = DkLangEtcUtils.sub_word([ta], fromWord="里", toWord="东西2")
    print(*ta3)

    # replace

    pass


if __name__ == "__main__":
    _f_main()


"""
a-z0-9A-Z_

  public static final String REGEX_CHINESE = "^[\u4e00-\u9fa5],{0,}$";




"""



import dknovautils as dku

from dknovautils.dkat import AT
from dknovautils.myadd import m_add


def test_ccc():

    print(AT.VERSION)

    assert m_add(1, 2) == 3


if __name__=="__main__":
    test_ccc()


"""

测试注释1

"""

# 单行注释1

from dknovautils import *

# from dknovautils.commons import *
# from dknovautils.dkfiles import DkPyFiles
# from dknovautils.dkat import AT

# from dknovautils.dkipy import DkIpyUtils

# from commons import *
# from dkfiles import DkPyFiles
# from dkat import AT

print("test01 start")

"""

测试注释2

"""

"""

测试注释3

"""


def test_d_001():
    """

    测试注释4

    What is a Circular Import?

    """

    assert 1 == 1, "err5162"

    assert str(None) == "None", str(None)

    assert AT.VERSION >= "0.0.39", f"err5484 {AT.VERSION}"

    AT.bindport_httpserver(port=10134, daemon=True)

    AT.log_basicConfig()

    AT.mybeep()

    AT.logger.info("hello dkn logger")

    time.sleep(1.0)

    # assert False, 'ok'

    iprint(f"now:{AT.fepochSecs()}")

    print(AT.sdf_isocompact_format_datetime())

    iprint_warn("hello22")
    iprint_info("hello22")
    iprint_error("hello22")

    assert AT._AstWarnCnt_ == 1 and AT._AstErrorCnt_ == 1

    print("=" * 5)

    # r=DkIpyUtils.mfr('dir')
    # iprint(r)

    # assert False,'ok'

    # return

    # AT.log_basicConfig()

    iprint(AT.VERSION)

    # AT.assert_(AT.is_windows)

    # fp = r'F:\DIKI\DIKIETC\ADEV\learn\LearnPython\PythonBook\dknovautils\test_a.py'

    # jsonstr = Path(fp).read_text(encoding='utf8')

    # 这是单行注释2
    # fp2 = DkPyFiles.f_remove_comments_v1(jsonstr)
    # # print(fp2)

    # iprint_info(fp2)

    cstr = AT.sdf_logger_format_datetime()
    print(cstr)

    def f_abc(a):
        return a

    # TypeError: cannot set 'f_abc' attribute of immutable type 'str'
    # str.f_abc = f_abc
    # assert "abc".f_abc(123)==123,'err9592'

    # 可以直接修改函数
    DkPyFiles.f_remove_comments_v1 = lambda x: x

    assert DkPyFiles.f_remove_comments_v1(123) == 123, "err2556"

    # now=datetime.now()
    # datetime.faa = 1

    # DkIpyUtils.magic_fc(123)

    assert DkIpyUtils.hello(1) == 1

    run_simple_a("ls")

    r = AT.md5_hex(b"abc")
    iprint(f"r: {r}")

    ra = AT.bindport_httpserver(port=8070)
    assert ra

    r = AT.bindport_httpserver(port=8070)
    assert not r

    ra.shutdown()

    x = 0

    def fxx():
        nonlocal x
        x += 1

    r = AT.bindport_httpserver_cb(port=8070, cb=fxx)
    r2 = AT.bindport_httpserver_cb(port=8070, cb=fxx)
    assert r and not r2 and x == 1
    r.shutdown()

    rc, out, err = run_simple_b(r"cmd.exe /c dir e:\etc ")
    iprint(out)

    if False:
        run_simple_a("rm /mnt/d/tmp/tmp/etc/* -rf")
        # 如此写法倒是可以勉强运行 但是中文都乱码了
        # robocopy的/unicode输出编码似乎有问题 搭配 'utf-16-le'还是不能正常运行的
        rc, out, err = run_simple_b(
            "robocopy.exe e:/etc d:/tmp/tmp/etc /MIR ",
            encoding="GBK",
            errors="replace",
        )
        iprint(out)

    if True:
        run_simple_a("rm /mnt/d/tmp/tmp/etc2/* -rf")
        # 如此写法似乎是最好的 中文的文件名称编码倒是正常的 只是有少部分中文乱码 
        # robocopy的/unicode输出编码似乎也有问题 搭配 'utf-16-le'还是不能正常运行的
        rc, out, err = run_simple_b(
            "robocopy.exe e:/etc2 d:/tmp/tmp/etc2 /MIR ",
            encoding="GBK",
            errors="replace",
        )
        iprint(out)

    print(AT.VERSION)


if __name__ == "__main__":
    test_d_001()


"""

https://packaging.python.org/en/latest/tutorials/packaging-projects/


"""



from __future__ import annotations


import http.server
from logging import Logger, LoggerAdapter
import socketserver
from typing import TypeVar
from typing_extensions import override


import dknovautils

# from dknovautils.dk_imports import *
from dknovautils.commons import *


DkAppVer = '0.2.9'

_unknown_err = "_unknown_err4035"

T = TypeVar("T")


class staticproperty(property):
    """
    https://stackoverflow.com/questions/1697501/staticmethod-with-property

    """

    def __get__(self, owner_self, owner_cls) -> Any:  # type:ignore
        return self.fget()  # type:ignore


class classproperty(property):
    def __get__(self, owner_self, owner_cls) -> Any:  # type:ignore
        return self.fget(owner_cls)  # type:ignore


"""
这是因为一个bug LoggerAdapter将在3.11版本中泛型化
https://github.com/python/typeshed/issues/7855
"""
if TYPE_CHECKING:
    MyLoggerAdapter = logging.LoggerAdapter[logging.Logger]
else:
    MyLoggerAdapter = logging.LoggerAdapter


# class MyLogger(LoggerAdapter):
#     pass


class AT(object):
    _inner_logger_fun = None

    _AstWarnCnt_: int = 0
    _AstErrorCnt_: int = 0
    _err_msgs: Deque[str] = deque(maxlen=100)

    _logger_adapter: MyLoggerAdapter | None = None
    _logger_name = "dkn"

    _is_windows: bool = sys.platform.startswith("win")
    _is_linux: bool = sys.platform.startswith("linux")

    @classproperty
    def is_windows(cls) -> bool:
        # r=sys.platform.startswith('win')
        r = cls._is_windows
        return r

    @classproperty
    def is_linux(cls) -> bool:
        # r=sys.platform.startswith('win')
        r = cls._is_linux
        return r

    @classmethod
    def _create_my_logger_adapter(cls) -> MyLoggerAdapter:
        class _MyLoggerAdapter(MyLoggerAdapter):
            # todo 这个 args kwargs 如何标注类型?
            @override
            def log(
                self, level: int, msg: object, *args, **kwargs  # type:ignore
            ) -> None:  # type:ignore
                if level == LLevel.Warn.value:
                    AT._AstWarnCnt_ += 1
                elif level >= LLevel.Error.value:
                    AT._AstErrorCnt_ += 1
                else:
                    pass

                super().log(level, msg, *args, **kwargs)

        log = logging.getLogger(cls._logger_name)
        r = _MyLoggerAdapter(logger=log)
        return r

    @classmethod
    def get_logger(cls, read_cache: bool = True) -> MyLoggerAdapter:
        if cls._logger_adapter is None:
            cls._logger_adapter = cls._create_my_logger_adapter()

        if read_cache:
            return cls._logger_adapter
        else:
            return cls._create_my_logger_adapter()

    @classproperty
    def logger(cls) -> MyLoggerAdapter:
        return cls.get_logger()

    @classmethod
    def log_loggerFun(
        cls,
        # loggerName: str = "dkn",
        initInnerLoggerFun: bool = True,
        beepError: bool = False,
        beepWarn: bool = False,
    ) -> Callable[[object, LLevel], None]:
        """
        创建一个logger 作为系统的缺省logger
        """

        def m_logger_fun(obj: object, llevel: LLevel) -> None:
            assert isinstance(llevel, LLevel)
            _logger = cls.get_logger()

            if False:
                pass
            elif llevel == LLevel.Trace:
                # 也用debug级别。只是在prod模式下可以关闭。
                _logger.debug(obj)
            elif llevel == LLevel.Debug:
                _logger.debug(obj)
            elif llevel == LLevel.Info:
                _logger.info(obj)
            elif llevel == LLevel.Warn:
                _logger.warning(obj)
                if beepWarn:
                    AT.beep_error_buffered()
            elif llevel == LLevel.Error:
                _logger.error(obj)
                if beepError:
                    AT.beep_error_buffered()
            else:
                assert False, f"bad {llevel}"

        if initInnerLoggerFun:
            cls._logger_adapter = None  # force recreate
            cls.get_logger()
            AT._inner_logger_fun = m_logger_fun

        return m_logger_fun

    _LOG_FORMAT_100 = "%(asctime)s - %(levelname)s - %(message)s"
    _DATE_FORMAT_100 = "%m/%d/%Y %H:%M:%S %p"
    # _DATE_FORMAT_iso_simple = "%m/%d/%Y %H:%M:%S %p"

    STRFMT_ISO_COMPACT_SIMPLE = "%Y%m%dT%H%M%S"
    STRFMT_ISO_COMPACT_ALL_A = "%Y%m%dT%H%M%SS%fZ%Z"
    STRFMT_ISO_ALL_A = "%Y-%m-%dT%H:%M:%S.%f%Z"
    STRFMT_ISO_SEC_A = "%Y-%m-%dT%H:%M:%S"

    STRFMT_LOGGER_MS_A = "%Y-%m-%d %H:%M:%S.%f"

    _default_time_format = "%Y-%m-%d %H:%M:%S"

    _LOG_FORMAT_106 = (
        "%(asctime)s.%(msecs)03d %(name)s %(threadName)s [%(levelname)s] %(message)s"
    )

    """
    
    dtstr = datetime.now().strftime("%Y%m%dT%H%M%S")

        
    """

    @classmethod
    def log_basicConfig(
        cls,
        filename: str | None = None,
        loggerName: str = "dkn",
        level: int = logging.DEBUG,
        initInnerLoggerFun: bool = True,
    ) -> Callable[[object, LLevel], None]:
        """
        这个 dkn logger的设置好像不太好。
        暂时不用这个

        """

        # AT.unsupported()

        logging.basicConfig(
            filename=filename,
            level=level,
            format=cls._LOG_FORMAT_106,
            datefmt=AT.STRFMT_ISO_SEC_A,
            # datefmt=cls._DATE_FORMAT_100,
            force=True,
        )

        logger = logging.getLogger(loggerName)
        logger.setLevel(level)

        # # create console handler and set level to debug
        # ch = logging.StreamHandler()
        # ch.setLevel(logging.DEBUG)

        # # create formatter
        # formatter = logging.Formatter(
        #     cls._LOG_FORMAT_106)

        # # add formatter to ch
        # ch.setFormatter(formatter)

        # # add ch to logger
        # logger.addHandler(ch)

        mloggerFun = AT.log_loggerFun(initInnerLoggerFun)

        return mloggerFun

    @staticmethod
    def assertAllNotNone(*args: Any) -> None:
        assert isinstance(args, tuple)
        AT.assert_(all(_ is not None for _ in args), "err7540 some value is None")

    @staticmethod
    def deg_to_rad(d: float) -> float:
        return d / 180.0 * math.pi

    __beep_last_time: float = 0

    @staticmethod
    def beep_error_buffered(tone: str = "ping") -> None:
        """
        在0.5秒内最多播放一次 类似防止按钮双击的效果

        """
        t = 0.5
        now = AT.fepochSecs()
        if now < AT.__beep_last_time + t:
            return
        else:
            AT.__beep_last_time = now
            AT.beep(tone)

    @staticmethod
    def beep(tone: str = "ping") -> None:
        if False:
            # import beepy  # type: ignore[import-untyped]
            # beepy.beep(sound=tone)
            pass

    @staticmethod
    def beep2() -> None:
        #!wget -q -T 1 http://localhost:333/hello
        print("\7")

    @classmethod
    def mybeep(
        cls,
        *,
        freq: float = 440,
        dur: float = 500,
        sleep: float | None = None,
        n: int = 1,
        lastDurRate: float = 1.0,
    ) -> None:
        if False:
            # 没用用到。只能播放特定wav文件。不适合动态生成信号。
            # import beepy
            # beepy.beep(sound=tone)
            pass

        if True and cls._is_windows:
            # from beep import beep
            import winsound

            freq = int(freq)
            dur = int(dur)

            for i in range(n):
                # beep(freq, dur)  # duration in ms, frequency in Hz
                r = 1.0 if i != n - 1 else lastDurRate
                dur2 = int(dur * r)
                winsound.Beep(freq, dur2)  # type: ignore
                if sleep is not None:
                    time.sleep(sleep)

        if True and cls._is_linux:
            freq = int(freq)

            for i in range(n):
                # beep -f 2000 -l 1500
                r = 1.0 if i != n - 1 else lastDurRate
                dur2 = int(dur * r)
                os.system(f"beep -f {freq} -l {dur2}")
                if sleep is not None:
                    time.sleep(sleep)

    @staticmethod
    def vassert_(b: bool, s: str | None = None) -> None:
        AT.assert_(b, s)

    @classmethod
    def never(cls, s: str | None = None) -> None:
        AT.assert_(False, s if s is not None else "should never come here")

    @classmethod
    def fail(cls, s: str | None = None) -> None:
        AT.assert_(False, s if s is not None else "expected failure err20384")

    @staticmethod
    def checksys() -> None:
        assert sys.version_info >= (3, 11)

    @classmethod
    def fepochMillis(cls) -> int:
        millisec = int(AT.fepochSecs() * 1000)
        return millisec

    _last_epochsecs = 0.0

    @classmethod
    def secs(cls) -> float:
        return cls.fepochSecs()

    @classmethod
    def fepochSecs(cls) -> float:
        """
        time.monotonic()不是epoch值!! F**K

        """
        r = time.time()
        if r <= cls._last_epochsecs:
            return cls._last_epochsecs
        else:
            cls._last_epochsecs = r
            return r

    """
        # dd/mm/YY H:M:S
    dts = now.strftime("%d/%m/%Y %H:%M:%S")
    print("date and time =", dts)

    # 20200101T120000
    dts = datetime.now().strftime("%Y%m%dT%H%M%S")
    print("date and time =", dts)
    
    
    """

    @classmethod
    def sdf_isocompact_format_datetime(
        cls, dt: datetime | None = None, *, precise: str = "s"
    ) -> str:
        """
        20201201T080100

        d: day
        s: second
        ms: millis
        a: all

        """
        assert precise in ["d", "s", "ms", "a"], "err5554 bad precise"

        AT.assert_(dt is None or isinstance(dt, datetime), "err95219")

        # if not isinstance(dt,datetime) and dt is not None:
        #     AT.unsupported()

        if dt is None:
            dt = AT._now_dt()

        dts = dt.strftime(AT.STRFMT_ISO_COMPACT_ALL_A)

        if precise == "a":
            pass
        elif precise == "ms":
            dts = dts[: (8 + 0 + 1 + 6 + 4)]
        elif precise == "s":
            dts = dts[: (8 + 0 + 1 + 6)]
        elif precise == "d":
            dts = dts[:(8)]
        else:
            AT.never()

        return dts

    @staticmethod
    def sdf_logger() -> None:
        AT.unimplemented()
        pass

    @staticmethod
    def _now_dt() -> datetime:
        """
        这个名称为何用下划线开头?

        """
        dt = datetime.now()
        return dt

    @classmethod
    def sdf_logger_format_datetime(
        cls,
        dt: int | datetime | None = None,
        *,
        precise: str = "s",
        noColon: bool = False,
    ) -> str:
        """
        https://strftime.org/
        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
        https://man7.org/linux/man-pages/man3/strftime.3.html

        -   增加一个参数 noColon，可以将冒号转换为-符号 这样的话 生成的结果可以直接用于文件名称中。

        2023-12-01T08:30:00.123456

        缺省返回
        "yyyy-MM-ddTHH:mm:ss"

        date -I

        d date
        s seconds
        ms millis
        a all

        """
        assert precise in ("d", "s", "ms", "a"), "err5554 bad precise"

        if dt is not None and isinstance(dt, int):
            dt = datetime.fromtimestamp(dt / 1000, tz=None)
        elif dt is None:
            dt = datetime.now()
        else:
            pass

        dts = dt.strftime(AT.STRFMT_ISO_ALL_A)

        if precise == "a":
            pass
        elif precise == "ms":
            dts = dts[: (10 + 1 + 8 + 4)]
        elif precise == "s":
            dts = dts[: (10 + 1 + 8)]
        elif precise == "d":
            dts = dts[:(10)]
        else:
            AT.never()

        if noColon:
            dts = dts.replace(":", "-")

        return dts

    @staticmethod
    def assert_(b: Any, s: str | None = None) -> None:
        # 改成完善的形式

        # assert isinstance(b, bool)

        if b:
            return

        # AT._AstErrorCnt_ += 1 后面的 iprint_error已经记录了错误 此处就不记录了
        msg = _unknown_err if s is None else s

        dknovautils.commons.iprint_error(msg)

        raise DkAstException(msg)

    @staticmethod
    def mychdir(s: str) -> None:
        assert isinstance(s, str) and len(s) > 0
        iprint_debug(f"chdir: {s}")
        os.chdir(s)

    @staticmethod
    def astTrace(b: bool, s: str) -> None:
        """在prod中完全不需要的"""
        AT.assert_(b, s)

    @staticmethod
    def unsupported(s: str = "feature") -> None:
        AT.assert_(False, f"err8255 unsupported [{s}]")

    @staticmethod
    def unimplemented(s: str = "feature") -> None:
        AT.assert_(False, f"err4173 unimplemented [{s}]")

    # @staticmethod
    # def f_matploglib_logger():
    #     logger = logging.getLogger('matplotlib.font_manager')
    #     logger.setLevel(level=logging.INFO)

    @staticmethod
    def log_conf_matploglib_logger(level: int = logging.INFO) -> None:
        logger = logging.getLogger("matplotlib.font_manager")
        logger.setLevel(level=level)

    @staticmethod
    def dict_group_value(d: Dict[Any, Any]) -> Dict[Any, Any]:
        d2 = {}
        for k, v in d.items():
            if not v in d2:
                d2[v] = [k]
            else:
                d2[v].append(k)
        return d2

    @classmethod
    def md5_hex(cls, bs: bytes) -> str:
        md5 = hashlib.md5()
        md5.update(bs)
        r = md5.hexdigest().lower()
        AT.assert_(len(r) == 32)
        return r

    @classmethod
    def get_file_md5_hex(cls, file_path: str | Path) -> str:
        """
        计算文件的md5
        :param file_name:
        :return:

        https://www.cnblogs.com/xiaodekaixin/p/11203857.html

        """
        m = hashlib.md5()  # 创建md5对象
        with open(file_path, "rb") as fobj:
            while True:
                data = fobj.read(4096)
                if not data:
                    break
                m.update(data)  # 更新md5对象

        return m.hexdigest().lower()  # 返回md5对象

    @classmethod
    def list_unique(cls, ls: List[T]) -> List[T]:
        d = {ln: None for ln in ls}
        lns = list(d.keys())
        return lns

    @classmethod
    def bindport_httpserver(
        cls, *, port: int, daemon: bool = False
    ) -> socketserver.TCPServer | None:
        """
        如果绑定端口成功 将启动一个线程 持续占用该端口 并返回 TcpServer
        否则返回 None

        """
        AT.assert_(port > 0, "err10591 port error")
        PORT = port
        Handler = http.server.SimpleHTTPRequestHandler
        server = None
        startOk = False

        try:

            def tf() -> None:
                nonlocal server
                nonlocal startOk
                try:
                    with socketserver.TCPServer(("", PORT), Handler) as httpd:
                        iprint_debug(f"serving at port {port}")
                        server = httpd
                        startOk = True
                        httpd.serve_forever()
                except:
                    startOk = False

            thread = threading.Thread(target=tf, args=(), daemon=daemon)
            thread.start()

            time.sleep(0.5)
            iprint_debug(f"bind tcp port ok. {port} {startOk}")
            return server

        except Exception as e:
            iprint_debug(f"some err occured: {e}")
            return None

    @classmethod
    def bindport_httpserver_cb(
        cls, *, port: int, cb: Callable[[], None]
    ) -> socketserver.TCPServer | None:
        """
        如果绑定端口成功 将启动一个daemon线程 持续占用该端口 然后执行cb回调函数. 返回 TCPServer
        如果绑定端口不成功 将直接 返回 None

        """
        AT.assert_(port > 0 and cb, "err27617")
        server = cls.bindport_httpserver(port=port, daemon=True)
        if server:
            try:
                cb()
            except Exception as e:
                iprint_warn(f"error78836 {e}")
            finally:
                return server
        else:
            return None

    # end
    VERSION = DkAppVer


class DkAstException(Exception):
    pass


"""

LOG_FORMAT = "%(asctime)s - %(levelname)s - %(message)s"
DATE_FORMAT = "%m/%d/%Y %H:%M:%S %p"

logging.basicConfig(filename='/mnt/d/tmp/demo0726.log', level=logging.DEBUG, format=LOG_FORMAT, datefmt=DATE_FORMAT)





"""
